
#ifndef TX_DWIN
   #ERROR You need to define the serial pins used for Dwin in globals.h
#endif

void ResetSignalRefreshTimer();
void DrawTitle();



#define DWIN_HEADER 0x5AA5
#define DWIN_CMD_WRITE 0x82

//Dwin VP addresses...
#define DWIN_TITLE1_VP_ADDRESS 0x1000   //title first line text box address
#define DWIN_TITLE2_VP_ADDRESS 0x1100   //title second line text box address

#define DWIN_TPMS_FL_PRESSURE_VP_ADDRESS 0x1200  //front left pressure text box
#define DWIN_TPMS_FR_PRESSURE_VP_ADDRESS 0x1300  //front right pressure text box
#define DWIN_TPMS_RL_PRESSURE_VP_ADDRESS 0x1400  //rear left pressure text box
#define DWIN_TPMS_RR_PRESSURE_VP_ADDRESS 0x1500  //rear right pressure text box


#define DWIN_TPMS_FL_TEMPERATURE_VP_ADDRESS 0x2200  //front left temperature text box
#define DWIN_TPMS_FR_TEMPERATURE_VP_ADDRESS 0x2300  //front right temperature text box
#define DWIN_TPMS_RL_TEMPERATURE_VP_ADDRESS 0x2400  //rear left temperature text box
#define DWIN_TPMS_RR_TEMPERATURE_VP_ADDRESS 0x2500  //rear right temperature text box


#define DWIN_TPMS_FL_PROGBAR_VP_ADDRESS 0x2200  //front left progress bar (indicating time since last update) min 0, max 5
#define DWIN_TPMS_FR_PROGBAR_VP_ADDRESS 0x2300  //front right progress bar (indicating time since last update) min 0, max 5
#define DWIN_TPMS_RL_PROGBAR_VP_ADDRESS 0x2400  //rear left progress bar (indicating time since last update) min 0, max 5
#define DWIN_TPMS_RR_PROGBAR_VP_ADDRESS 0x2500  //rear right progress bar (indicating time since last update) min 0, max 5


enum VPTYPES
{
  VP_PRESSURE,
  VP_TEMPERATURE,
  VP_PROGBAR
};

#include "SoftwareSerial.h"


SoftwareSerial Serial_Dwin(DWIN_RX, DWIN_TX);

void DisplayInit()
{
  Serial_Dwin.begin(9600); 

}

void ScreenSetup()
{
    DisplayInit();

    //add any screen initailisations here (e.g. clear text boxes etc.

    DrawTitle();  
}

void Dwin_SendText(char* txt, unsigned int VP)
{
   Serial_Dwin.write(highByte(DWIN_HEADER));
   Serial_Dwin.write(lowByte(DWIN_HEADER));
   Serial_Dwin.write((byte)(strlen(txt) + 3)));
   Serial_Dwin.write(DWIN_CMD_WRITE);
   Serial_Dwin.write(highByte(VP));
   Serial_Dwin.write(lowByte(VP));
   Serial_Dwin.write(txt, strlen(txt)); 
   
}

void Dwin_SendInt(int val, unsigned int VP)
{
   Serial_Dwin.write(highByte(DWIN_HEADER));
   Serial_Dwin.write(lowByte(DWIN_HEADER));
   Serial_Dwin.write(5);
   Serial_Dwin.write(DWIN_CMD_WRITE);
   Serial_Dwin.write(highByte(VP));
   Serial_Dwin.write(lowByte(VP));
   Serial_Dwin.write(highByte(val));
   Serial_Dwin.write(lowByte(val));
   
}

void DrawTitle()
{
    
    #ifdef NissanLeaf
       Dwin_SendText(F("Nissan Leaf TPMS"),DWIN_TITLE1_VP_ADDRESS);
    #elif Dacia
       Dwin_SendText(F("Dacia TPMS"),DWIN_TITLE1_VP_ADDRESS);
    #elif Renault
       #ifdef Zoe
          Dwin_SendText(F("Ren Zoe(early) TPMS"),DWIN_TITLE1_VP_ADDRESS);
       #else
          Dwin_SendText(F("Renault TPMS"),DWIN_TITLE1_VP_ADDRESS);
       #endif
    #elif Citroen
       Dwin_SendText(F("Citroen TPMS"),DWIN_TITLE1_VP_ADDRESS);
    #elif Jansite
       Dwin_SendText(F("Jansite TPMS"),DWIN_TITLE1_VP_ADDRESS);
    #elif JansiteSolar
       Dwin_SendText(F("JSolar TPMS"),DWIN_TITLE1_VP_ADDRESS);
    #elif Ford
       Dwin_SendText(F("Ford TPMS"),DWIN_TITLE1_VP_ADDRESS);
    #elif PontiacG82009
       Dwin_SendText(F("Pontiac G8 TPMS"),DWIN_TITLE1_VP_ADDRESS);
    #elif Hyundai_i35
       Dwin_SendText(F("Hyundai i35"),DWIN_TITLE1_VP_ADDRESS);
    #elif Schrader_C1100
       Dwin_SendText(F("Hyundai Tucson"),DWIN_TITLE1_VP_ADDRESS);
    #elif Schrader_A9054100
       Dwin_SendText(F("Smart ForTwo TPMS"),DWIN_TITLE1_VP_ADDRESS); 
    #elif Subaru
       Dwin_SendText(F("Subaru TPMS"),DWIN_TITLE1_VP_ADDRESS);  
    #else
       Dwin_SendText(F("Toyota TPMS"),DWIN_TITLE1_VP_ADDRESS);
    #endif
    char Mess[100];
    sprintf(Mess,"JSM Solutions ");
    sprintf(Mess + strlen(Mess),VERSION);
    Dwin_SendText(Mess,DWIN_TITLE2_VP_ADDRESS);

    //display.println(")");
  
  
  }

  char DisplayTimeoutBar(uint32_t TimeSinceLastUpdate)
  {
      int16_t HowCloseToTimeout;
      HowCloseToTimeout = (int16_t)(TimeSinceLastUpdate/(TPMS_TIMEOUT/5));

      switch(HowCloseToTimeout)
      {
        case 0: 
           //return(FONTBAR_7);
           return('5');
           break;
        case 1: 
           //return(FONTBAR_5);
           return('4');
           break;
        case 2: 
           //return(FONTBAR_3);
           return('3');
           break;
        case 3: 
           //return(FONTBAR_2);
           return('2');
           break;
        case 4: 
           //return(FONTBAR_1);
           return('1');
           break;
        default: 
           //return(FONTBAR_0);
           return('0');
           break;
                      
      }
  }

  unsigned int GetVP(int16_t DisplayIndex, VPTYPES VPType)
  {
     switch (DisplayIndex)
     {
        case 0: 
           switch(VPType)
           {
               case VP_PRESSURE:
                  return(DWIN_TPMS_FL_PRESSURE_VP_ADDRESS);
               break;
               case VP_TEMPERATURE:
                  return(DWIN_TPMS_FL_TEMPERATURE_VP_ADDRESS);
               break;
               case VP_PROGBAR:
                  return(DWIN_TPMS_FL_PROGBAR_VP_ADDRESS);
               break;
           }
           break;
           
        case 1: 
           switch(VPType)
           {
               case VP_PRESSURE:
                  return(DWIN_TPMS_FR_PRESSURE_VP_ADDRESS);
               break;
               case VP_TEMPERATURE:
                  return(DWIN_TPMS_FR_TEMPERATURE_VP_ADDRESS);
               break;
               case VP_PROGBAR:
                  return(DWIN_TPMS_FR_PROGBAR_VP_ADDRESS);
               break;
           }
           break;
        case 2: 
           switch(VPType)
           {
               case VP_PRESSURE:
                  return(DWIN_TPMS_RL_PRESSURE_VP_ADDRESS);
               break;
               case VP_TEMPERATURE:
                  return(DWIN_TPMS_RL_TEMPERATURE_VP_ADDRESS);
               break;
               case VP_PROGBAR:
                  return(DWIN_TPMS_RL_PROGBAR_VP_ADDRESS);
               break;
           }
           break;
        case 3: 
           switch(VPType)
           {
               case VP_PRESSURE:
                  return(DWIN_TPMS_RR_PRESSURE_VP_ADDRESS);
               break;
               case VP_TEMPERATURE:
                  return(DWIN_TPMS_RR_TEMPERATURE_VP_ADDRESS);
               break;
               case VP_PROGBAR:
                  return(DWIN_TPMS_RR_PROGBAR_VP_ADDRESS);
               break;
           }
           break;
     }
     return(0);
  }


  void UpdateBlock(int16_t DisplayIndex,int16_t i)
  {
        int16_t x,y;
        char s[25];
        char t;

        if ((TPMS[i].LowPressure == true) || (TPMS[i].HighPressure == true))
        {
          if (DisplayFlash)
          {
             strcpy(s,"    ");
          }
          else
          {
             #ifdef DISPLAY_PRESSURE_AS_BAR
                dtostrf(PSI_To_BAR(TPMS[i].TPMS_Pressure), 4, 2, s);
             #elif DISPLAY_PRESSURE_AS_KPA
                dtostrf(PSI_To_KPA(TPMS[i].TPMS_Pressure), 3, 0, s);  //rounded to integer value
             #else
                dtostrf(TPMS[i].TPMS_Pressure, 3, 1, s);
             #endif
          }
        }
        else
        {
           #ifdef DISPLAY_PRESSURE_AS_BAR
              dtostrf(PSI_To_BAR(TPMS[i].TPMS_Pressure), 4, 2, s);
           #elif DISPLAY_PRESSURE_AS_KPA
              dtostrf(PSI_To_KPA(TPMS[i].TPMS_Pressure), 3, 0, s);  //rounded to integer value
           #else
              dtostrf(TPMS[i].TPMS_Pressure, 3, 1, s);
           #endif
        }

        //pressure
        Dwin_SendText(s,GetVP(DisplayIndex,VPTYPES.PRESSURE));

        #ifdef DISPLAY_TEMP_AS_FAHRENHEIT
           t = 'F';
        #else
           t = 'C';
        #endif

        if (TPMS[i].TPMS_Temperature == NO_VALID_TEMPERATURE)
        {
           Dwin_SendText("  ---  ",GetVP(DisplayIndex,VPTYPES.TEMPERATURE));
        }
        else
        {
  
          if ((TPMS[i].LowTemperature == true) || (TPMS[i].HighTemperature == true))
          {
            if (DisplayFlash)
            {
               strcpy(s,"       ");
               Dwin_SendText(s,GetVP(DisplayIndex,VPTYPES.TEMPERATURE));
            }
            else
            {
               #ifdef DISPLAY_TEMP_AS_FAHRENHEIT
                  dtostrf(DegC_To_DegF(TPMS[i].TPMS_Temperature), 2, 0, s);
               #else
                  dtostrf(TPMS[i].TPMS_Temperature, 2, 0, s);
               #endif

               strcat(s," ");  //add a space
               strcat(s,t);  //append the F or C
               Dwin_SendText(s,GetVP(DisplayIndex,VPTYPES.TEMPERATURE));
            }
          }
          else
          {
               #ifdef DISPLAY_TEMP_AS_FAHRENHEIT
                  dtostrf(DegC_To_DegF(TPMS[i].TPMS_Temperature), 2, 0, s);
               #else
                  dtostrf(TPMS[i].TPMS_Temperature, 2, 0, s);
               #endif
               strcat(s," ");  //add a space
               strcat(s,t);  //append the F or C
               Dwin_SendText(s,GetVP(DisplayIndex,VPTYPES.TEMPERATURE));
          }
        }



        //display vertical bar showing how long since last update 7 bars = recent 1 bar = nearing timeout (at timeout it will be removed from display altogether)
        int val =DisplayTimeoutBar(millis() - TPMS[i].lastupdated);
        Dwin_SendInt(val,GetVP(DisplayIndex,VPTYPES.PROGBAR)

  }
 
  void UpdateDisplay()
  {
    int16_t i;

  
    int16_t DisplayIndex;
  

    //for (i = 0; i < 4; i++)
    for (i = 0; i < TYRECOUNT; i++) 
    {
      
      if (TPMS[i].TPMS_ID != 0)
      {
         //Only update the areas which need it to keep the timing overheads down
         DisplayIndex = GetDisplayIndexFromTyreIndex(i);

         if (DisplayIndex >= 0)
         {
           UpdateBlock(DisplayIndex,i);
           if ((bitRead(TPMSChangeBits,i) == 1))
           {
              bitClear(TPMSChangeBits,i);
           }
         }

      }


    }    
  
  }
